@@ -18,23 +18,55 @@ from socket import (
 )
 from threading import Thread, Timer
 from time import sleep
+from typing import Dict, Optional, Set
 from wsgiref import simple_server
 
 from werkzeug.serving import WSGIRequestHandler
 
 from .gating import wait_for_idle_cpus
+from .oob_response import OutOfBandResponder
 from .resource_balancer import BoundedResourceBalancer, ResourceLoan
+from .util import ObjectProxy
 
 LOGGER = getLogger("pothead.worker")
+BROKER_CONNECT_RETRY_INTERVAL = 2
+
+
+class DeferrableCleanup:
+    def __init__(self, f):
+        self.f = f
+
+    def assume(self):
+        f = self.f
+        self.f = None
+        return f
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, trace):
+        if self.f:
+            self.f()
 
 
 class Handler(WSGIRequestHandler):
     protocol_version = "HTTP/1.1"
 
-    def __init__(self, socket, address, server, cb_request_received):
+    def __init__(
+        self, socket, address, server, cb_request_received, cb_processing_done
+    ):
         self.cb_request_received = cb_request_received
         self.env = None
-        super().__init__(socket, address, server)
+
+        with DeferrableCleanup(cb_processing_done) as c:
+            if isinstance(server.app, OutOfBandResponder):
+                app = server.app
+                server = ObjectProxy(server)
+                server.app = lambda environ, start_response: app(
+                    environ, start_response, c.assume
+                )
+
+            super().__init__(socket, address, server)
 
     def handle_one_request(self):
         try:
@@ -104,48 +136,62 @@ class Server:
     ssl_context = None
     multithread = False
     multiprocess = False
-    server_address = "localhost"
     passthrough_errors = False
     shutdown_signal = False
     running = True
 
-    def __init__(self, addr, app, load_balancer=LoadBalancer):
+    def __init__(
+        self,
+        addr,
+        app,
+        *,
+        load_balancer=LoadBalancer,
+        redirect_port: Optional[int] = None
+    ):
         (host, port) = addr
         # Set up base environment
-        env = self.base_environ = {}
+        env: Dict[str, str] = {}
         env["SERVER_NAME"] = host
         env["GATEWAY_INTERFACE"] = "HTTP/1.1"
         env["SERVER_PORT"] = port
         env["REMOTE_HOST"] = ""
         env["CONTENT_LENGTH"] = ""
         env["SCRIPT_NAME"] = ""
+        self.base_environ = env
 
         self.broker_balancer = load_balancer(host, port)
-        self.app = app
-        self.sockets_waiting_for_request = set()
+        self.sockets_waiting_for_request: Set[socket] = set()
+        self.real_app = app
+        self.app = (
+            OutOfBandResponder(app, redirect_port) if redirect_port is not None else app
+        )
 
     def worker(self):
         while self.running:
             with self.broker_balancer.acquire_addr() as addr:
                 self._fetch_and_run_one_job(addr)
 
-    def _fetch_and_run_one_job(self, addr, cb_request_received=None):
+    def _fetch_and_run_one_job(
+        self, addr, cb_request_received=None, cb_processing_done=lambda _: None
+    ):
         with socket(AF_INET, SOCK_STREAM) as s:
             self.sockets_waiting_for_request.add(s)
             try:
                 s.connect(addr)
+                self.server_address = s.getsockname()
 
                 s.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
                 s.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
                 s.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 2)
                 s.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 3)
             except Exception as e:
-                LOGGER.error("Connect To Broker: %s", e)
-                sleep(2)
+                LOGGER.error("Couldn't connect to broker: %r", e)
+                cb_processing_done()
+                sleep(BROKER_CONNECT_RETRY_INTERVAL)
                 return
 
             try:
-                Handler(s, addr, self, cb_request_received)
+                Handler(s, addr, self, cb_request_received, cb_processing_done)
             except BaseException:
                 LOGGER.exception("In WSGI request handling")
                 sleep(1)
@@ -158,15 +204,18 @@ class Server:
         LOGGER.info("shutting down workers...")
         self.running = False
         for s in list(self.sockets_waiting_for_request):
-            s.shutdown(SHUT_RDWR)
-            s.close()
+            try:
+                s.shutdown(SHUT_RDWR)
+                s.close()
+            except OSError:
+                pass
 
     def poll_loop(self):
         """Uses app.wait_for_slot() spawn new workers dynamically. This is useful, for example to wait for
         enough CPU, memory, or disk space, to start next job.
 
         `wait_for_slot` is expected to block until it's time to poll for a new job. It is given a function
-        `halt` which it should polle intermittently to check if this worker is undergoing `shutdown()`.
+        `halt` which it should poll intermittently to check if this worker is undergoing `shutdown()`.
 
         It may return an object, if so `obj.request_received(environ)` will be called when a request is
         received and `obj.done()` when either the request is received, or if polling failed in which case
@@ -175,7 +224,7 @@ class Server:
         See pothead.gating for some ready-made `wait_for_slot` implementations"""
 
         while self.running:
-            callbacks = self.app.wait_for_slot(lambda: not self.running)
+            callbacks = self.real_app.wait_for_slot(lambda: not self.running)
             if not self.running:
                 break
 
@@ -191,12 +240,13 @@ class Server:
             broker_instance_lease.release()
             getattr(callbacks, "request_received", lambda _: None)(environ)
 
-        try:
-            with broker_instance_lease as addr:
-                self._fetch_and_run_one_job(addr, request_received)
-        finally:
-            if hasattr(callbacks, "done"):
-                callbacks.done()
+        with broker_instance_lease as addr:
+            self._fetch_and_run_one_job(
+                addr, request_received, getattr(callbacks, "done", lambda _: None)
+            )
+
+    def log(self, type, msg, *args, **kwargs):
+        getattr(LOGGER, type)(msg.rstrip(), *args, **kwargs)
 
 
 def install_term_handler(f):
@@ -243,6 +293,11 @@ if __name__ == "__main__":
         type=address,
         help="Load Balancer Hub to connect to [host:port]",
     )
+    parser.add_argument(
+        "--redirect-response",
+        default=None,
+        help="When streaming successful responses, first redirect client to worker at [port]",
+    )
     job_control_group = parser.add_mutually_exclusive_group()
     job_control_group.add_argument(
         "--workers",
@@ -268,7 +323,7 @@ if __name__ == "__main__":
 
     LOGGER.info("Initializing server for app %s", args.app)
 
-    server = Server(args.connect, args.app)
+    server = Server(args.connect, args.app, redirect_port=args.redirect_response)
 
     install_term_handler(server.shutdown)
 
